"""
Tests for L{documint.util}.
"""
from StringIO import StringIO

from lxml import etree
from twisted.trial.unittest import TestCase

from documint.util import embedStylesheets



class EmbedStylesheetsTests(TestCase):
    """
    Tests for L{documint.util.embedStylesheets}.
    """
    def test_removeStylesheetLinks(self):
        """
        L{documint.util.embedStylesheets} removes links with the
        C{'stylesheet'} relationship.
        """
        markup = embedStylesheets('''
<!DOCTYPE html PUBLIC "-//W3C//DTD XHTML 1.0 Strict//EN" "http://www.w3.org/TR/xhtml1/DTD/xhtml1-strict.dtd">
<html xml:lang="en" xmlns="http://www.w3.org/1999/xhtml"><head>
    <link href="test.css" type="text/css" rel="stylesheet" />
    <link type="text/css" rel="something" />
</head><body></body></html>''', [], removeStylesheets=True)
        tree = etree.parse(StringIO(markup))
        namespaces = {'xhtml': 'http://www.w3.org/1999/xhtml'}
        elems = tree.findall('//xhtml:head/xhtml:link', namespaces=namespaces)
        self.assertEquals(1, len(elems))
        self.assertEquals('something', elems[0].get('rel'))


    def test_embedStylesheets(self):
        """
        L{documint.util.embedStylesheets} embeds stylesheets in I{style}
        elements.
        """
        stylesheets = [
            'div {color:red;}',
            'span {color:blue;}']
        markup = embedStylesheets('''
<!DOCTYPE html PUBLIC "-//W3C//DTD XHTML 1.0 Strict//EN" "http://www.w3.org/TR/xhtml1/DTD/xhtml1-strict.dtd">
<html xml:lang="en" xmlns="http://www.w3.org/1999/xhtml"><head></head><body></body></html>''', stylesheets)
        tree = etree.parse(StringIO(markup))
        namespaces = {'xhtml': 'http://www.w3.org/1999/xhtml'}
        elems = tree.findall('//xhtml:head/xhtml:style', namespaces=namespaces)
        self.assertEquals(2, len(elems))
        for elem, stylesheet in zip(elems, stylesheets):
            self.assertEquals(stylesheet, elem.text)



"""
I{Documint} I{AMP} client.
"""
from twisted.internet import task
from twisted.internet.endpoints import clientFromString, connectProtocol
from twisted.protocols.amp import AMP

from documint.commands import Render



def render(protocol, markup, stylesheets):
    """
    Execute the L{Render} command on an AMP protocol instance.
    """
    return protocol.callRemote(
        Render,
        markup=markup,
        stylesheets=stylesheets)



__all__ = ['render']



if __name__ == '__main__':
    import sys
    def main(reactor, markup, styles):
        def _readFile(name):
            with file(name, 'rb') as fd:
                return fd.read()

        def _writeResponse(response):
            sys.stdout.write(response['data'])
            sys.stdout.flush()

        endpoint = clientFromString(reactor, 'tcp:host=127.0.0.1:port=8750')
        d = connectProtocol(endpoint, AMP())
        d.addCallback(render, _readFile(markup), map(_readFile, styles))
        d.addCallback(_writeResponse)
        return d

    task.react(main, (sys.argv[1], sys.argv[2:]))



"""
Documint I{AMP} command and protocol definitions.
"""
from lxml import etree
from twisted.internet.defer import succeed
from twisted.protocols import amp
from twisted.python.failure import Failure

from documint.errors import (
    ExternalProcessError, RemoteExternalProcessError, XMLSyntaxError)
from documint.extproc.css2xslfo import renderXHTML
from documint.mediumbox import BigString
from documint.util import embedStylesheets



class Render(amp.Command):
    """
    Render I{XHTML} markup and I{CSS} documents to a I{PDF} document.

    Accepts the following arguments:

    * C{markup}, a L{str}, that is the I{XHTML} markup byte data;
    * C{stylesheets}, a L{list} of L{str}, that is a list of stylesheet byte
      data.

    Returns a L{dict} mapping C{'data'} to a C{str} of rendered byte data, and
    C{'contentType'} to the content type of the byte data.
    """
    arguments = [
        ('markup', BigString()),
        ('stylesheets', amp.ListOf(amp.String()))]

    response = [
        ('data', BigString()),
        ('contentType', amp.String())]

    errors = {
        XMLSyntaxError: 'XML_SYNTAX_ERROR',
        RemoteExternalProcessError: 'EXTERNAL_PROCESS_ERROR'}



class Minter(amp.AMP):
    """
    Documint AMP protocol.
    """
    def embedStylesheets(self, markup, stylesheets):
        """
        Embed external stylesheets in I{XHTML} markup.

        @type  markup: L{str}
        @param markup: I{XHTML} markup byte data.

        @type  stylesheets: I{iterable} of L{str}
        @param stylesheets: Iterable of stylesheet byte data to embed.

        @rtype: L{str}
        @return: I{XHTML} markup with embedded stylesheets.
        """
        try:
            return embedStylesheets(markup, stylesheets)
        except etree.XMLSyntaxError, e:
            raise XMLSyntaxError(e)


    def renderXHTML(self, markup, stylesheets):
        """
        Render I{XHMTL} markup and I{CSS} to a I{PDF}.

        @type  markup: L{str}
        @param markup: I{XHTML} UTF-8 byte data.

        @type  stylesheets: I{iterable} of L{str}
        @param stylesheets: Iterable of stylesheet UTF-8 byte data to embed.

        @rtype: L{Deferred} firing with C{(str, str)}
        @return: Deferred that fires with the generated I{PDF} byte data and
            content type.
        """
        def _handleExternalProcessError(f):
            f.trap(ExternalProcessError)
            return Failure(RemoteExternalProcessError(f.getErrorMessage()))
        d = renderXHTML(self.embedStylesheets(markup, stylesheets))
        d.addErrback(_handleExternalProcessError)
        d.addCallback(lambda data: (data, 'application/pdf'))
        return d


    @Render.responder
    def render(self, markup, stylesheets):
        d = self.renderXHTML(markup, stylesheets)
        d.addCallback(
            lambda (data, contentType):
                dict(data=data,
                     contentType=contentType))
        return d



class SimpleMinter(Minter):
    """
    Simple Documint AMP protocol.

    The main difference between L{SimpleMinter} and L{Minter} is that
    L{SimpleMinter} will output I{XHTML}.
    """
    def renderXHTML(self, markup, stylesheets):
        """
        Embed I{CSS} in I{XHMTL} markup

        @type  markup: L{str}
        @param markup: I{XHTML} UTF-8 byte data.

        @type  stylesheets: I{iterable} of L{str}
        @param stylesheets: Iterable of stylesheet UTF-8 byte data to embed.

        @rtype: L{Deferred} firing with C{(str, str)}
        @return: Deferred that fires with the generated I{XHTML}, including
            embedded I{CSS} byte data, and the content type.
        """
        return succeed(
            (self.embedStylesheets(markup, stylesheets), 'text/html'))



__all__ = ['Render', 'Minter', 'SimpleMinter']



"""
I{Documint} error types.
"""



class XMLSyntaxError(Exception):
    """
    Wrapper around L{lxml.etree.XMLSyntaxError} that requires no additional
    arguments.
    """


class ExternalProcessError(RuntimeError):
    """
    An external process returned an exit status indicating failure.

    @type binary: C{str}
    @ivar binary: Path to the binary to spawn.

    @type arguments: C{sequence} of C{str}
    @ivar arguments: Arguments to pass when spawning L{binary}.

    @type code: C{int}
    @ivar code: Exit status.

    @type stdout: C{str}
    @ivar stdout: Standard output data.

    @type stderr: C{str}
    @ivar stderr: Standard error data.
    """
    def __init__(self, binary, arguments, code, (stdout, stderr)):
        RuntimeError.__init__(self, (binary, arguments, code, (stdout, stderr)))
        self.binary = binary
        self.arguments = arguments
        self.code = code
        self.stdout = stdout
        self.stderr = stderr



class RemoteExternalProcessError(Exception):
    """
    AMP-friendly description of an external process error.
    """



class NoSuchFile(IOError):
    """
    The specified file could not be found.
    """



import tempfile
from functools import partial
from os.path import expanduser

from twisted.python.filepath import FilePath

from documint.extproc.common import getProcessOutput, which
from documint.extproc.fop import fop
from documint.util import defertee



FOP_CONFIG = FilePath(expanduser('~/.config/documint/fop.xconf'))
findCSS2XSLFO = partial(which, 'css2xslfo')



def css2xslfo(xhtmlPath, xslfoPath):
    """
    Invoke I{CSS2XSLFO}.

    @type  xhtmlPath: L{FilePath}
    @param xhtmlPath: Path to read I{XHTML} input from.

    @type  xslfoPath: L{FilePath}
    @param xslfoPath: Path to write I{XSL-FO} output to.

    @raises ExternalProcessError: If I{CSS2XSLFO} did not exit successfully.

    @rtype: C{Deferred}
    @return: Deferred that fires when the process is complete.
    """
    def checkStatus((out, err, code)):
        return xslfoPath.getsize() > 0

    return getProcessOutput(
        findCSS2XSLFO(),
        ['-fo', xslfoPath.path,
         xhtmlPath.path],
        checkStatus=checkStatus)



def renderXHTML(markup, tempDir=None, css2xslfo=css2xslfo, fop=fop):
    """
    Render an I{XHTML} document to a I{PDF} document.

    @type  markup: L{str}
    @param markup: I{XHTML} document, encoded as UTF-8, that includes
        stylesheet information.

    @rtype: L{Deferred} firing with L{str}
    @return: Deferred that fires with the generated I{PDF} byte data.
    """
    def _removeTemp(ignored):
        tempDir.remove()

    if tempDir is None:
        tempDir = FilePath(tempfile.mkdtemp())
    xhtmlPath = tempDir.child('input.html')
    xhtmlPath.setContent(markup)
    xslfoPath = tempDir.child('output.fo')
    pdfPath = tempDir.child('output.pdf')

    configPath = FOP_CONFIG
    if not configPath.exists():
        configPath = None

    d = css2xslfo(xhtmlPath, xslfoPath)
    d.addCallback(lambda ignored: fop(xslfoPath, pdfPath, configPath))
    d.addCallback(defertee, _removeTemp)
    return d


__all__ = ['css2xslfo', 'renderXHTML']



from twisted.application import strports
from twisted.application.service import IServiceMaker
from twisted.internet import reactor
from twisted.internet.protocol import Factory
from twisted.plugin import IPlugin
from twisted.python import usage
from zope.interface import implements

from documint.commands import Minter, SimpleMinter



class Options(usage.Options):
    optFlags = [
        ['simple', 's', 'Simple Documint service']]

    optParameters = [
        ['port', 'p', 'tcp:8750', 'Documint service strport description']]



class DocumintServiceMaker(object):
    implements(IServiceMaker, IPlugin)
    tapname = 'documint'
    description = 'Document creation service'
    options = Options


    def makeService(self, options):
        factory = Factory()
        if options['simple']:
            factory.protocol = SimpleMinter
        else:
            factory.protocol = Minter
        return strports.service(options['port'], factory, reactor=reactor)


serviceMaker = DocumintServiceMaker()



from twisted.internet.defer import fail, succeed
from twisted.python.filepath import FilePath
from twisted.trial.unittest import TestCase

from documint.errors import ExternalProcessError
from documint.extproc.css2xslfo import css2xslfo, findCSS2XSLFO, renderXHTML



class CSS2XSLFOTests(TestCase):
    """
    Tests for L{documint.util.css2xslfo}.
    """
    try:
        findCSS2XSLFO()
    except RuntimeError:
        skip = 'css2xslfo unavailable'


    def setUp(self):
        self.dataPath = FilePath(__file__).sibling('data')


    def test_missing(self):
        """
        Invoking I{css2xslfo} with a missing input raises an
        L{ExternalProcessError}.
        """
        def checkException(e):
            self.assertIn('(No such file or directory)', e.stderr)

        outputPath = FilePath(self.mktemp())
        outputPath.touch()
        d = css2xslfo(self.dataPath.child('missing.html'), outputPath)
        d = self.assertFailure(d, ExternalProcessError)
        d.addCallback(checkException)
        return d


    def test_broken(self):
        """
        Invoking I{css2xslfo} with a broken input raises an
        L{ExternalProcessError}.
        """
        def checkException(e):
            self.assertIn(
                'element type "div" must be terminated by the matching end-tag',
                e.stderr)
        outputPath = FilePath(self.mktemp())
        outputPath.touch()
        d = css2xslfo(self.dataPath.child('broken.html'), outputPath)
        d = self.assertFailure(d, ExternalProcessError)
        d.addCallback(checkException)
        return d



class RenderXHMTLTests(TestCase):
    """
    Tests for L{documint.util.renderXHTML}.
    """
    def mkdtemp(self):
        """
        Create a temporary directory.

        @rtype: L{FilePath}
        """
        tempDir = FilePath(self.mktemp())
        if not tempDir.exists():
            tempDir.makedirs()
        return tempDir


    def test_renderXHTML(self):
        """
        L{renderXHTML} invokes L{css2xslfo} then L{fop}, removes the temporary
        directory, and returns I{fop}'s result.
        """
        def mockCSS2XSLFO(xhtmlPath, xslfoPath):
            return succeed(None)

        def mockFop(xslfoPath, pdfPath, configFile=None):
            return succeed('pdf')

        def cb(pdfData):
            self.assertIdentical(str, type(pdfData))
            self.assertEquals('pdf', pdfData)
            self.assertFalse(tempDir.exists())

        tempDir = self.mkdtemp()
        d = renderXHTML(
            'markup',
            tempDir=tempDir,
            css2xslfo=mockCSS2XSLFO,
            fop=mockFop)
        d.addCallback(cb)
        return d


    def test_renderXHTMLCSS2XSLFOFails(self):
        """
        If L{renderXHTML} fails invoking L{css2xslfo}, L{fop} is not invoked
        and the temporary directory is not removed.
        """
        def mockCSS2XSLFO(xslfoPath, xhtmlPath):
            return fail(RuntimeError(1))

        def mockFop(xslfoPath, pdfPath, configFile=None):
            self.fail('Never get here')

        def cb(e):
            self.assertEquals('1', str(e))
            self.assertTrue(tempDir.exists())

        tempDir = self.mkdtemp()
        d = renderXHTML(
            'markup',
            tempDir=tempDir,
            css2xslfo=mockCSS2XSLFO,
            fop=mockFop)
        d = self.assertFailure(d, RuntimeError)
        d.addCallback(cb)
        return d


    def test_renderXHTMLFopFails(self):
        """
        If L{renderXHTML} fails invoking L{fop}, the temporary
        directory is not removed.
        """
        def mockCSS2XSLFO(xhtmlPath, xslfoPath):
            return succeed(None)

        def mockFop(xslfoPath, pdfPath, configFile=None):
            return fail(RuntimeError(2))

        def cb(e):
            self.assertEquals('2', str(e))
            self.assertTrue(tempDir.exists())

        tempDir = self.mkdtemp()
        d = renderXHTML(
            'markup',
            tempDir=tempDir,
            css2xslfo=mockCSS2XSLFO,
            fop=mockFop)
        self.assertFailure(d, RuntimeError)
        d.addCallback(cb)
        return d



# From lp:~glyph/+junk/amphacks
"""
An argument type for sending medium-sized strings (more than 64k, but small
enough that they still fit into memory and don't require streaming).
"""

from cStringIO import StringIO
from itertools import count

from twisted.protocols.amp import AMP, Argument, Command



CHUNK_MAX = 0xffff

class BigString(Argument):
    def fromBox(self, name, strings, objects, proto):
        value = StringIO()
        value.write(strings.get(name))
        for counter in count(2):
            chunk = strings.get("%s.%d" % (name, counter))
            if chunk is None:
                break
            value.write(chunk)
        objects[name] = value.getvalue()


    def toBox(self, name, strings, objects, proto):
        value = StringIO(objects[name])
        firstChunk = value.read(CHUNK_MAX)
        strings[name] = firstChunk
        counter = 2
        while True:
            nextChunk = value.read(CHUNK_MAX)
            if not nextChunk:
                break
            strings["%s.%d" % (name, counter)] = nextChunk
            counter += 1

class Send(Command):
    arguments = [('big', BigString())]

class Example(AMP):
    @Send.responder
    def gotBig(self, big):
        print 'Got a big input', len(big)
        f = file("OUTPUT", "wb")
        f.write(big)
        f.close()
        return {}

def main(argv):
    from twisted.internet import reactor
    from twisted.internet.protocol import Factory, ClientCreator
    if argv[1] == 'client':
        filename = argv[2]
        def connected(result):
            result.callRemote(Send, big=file(filename).read())
        ClientCreator(reactor, AMP).connectTCP("localhost", 4321).addCallback(
            connected)
        reactor.run()
    elif argv[1] == 'server':
        f = Factory()
        f.protocol = Example
        reactor.listenTCP(4321, f)
        reactor.run()
    else:
        print "Specify 'client' or 'server'."
if __name__ == '__main__':
    from sys import argv as arguments
    main(arguments)



from os.path import exists
from warnings import warn

from twisted.internet.utils import getProcessOutputAndValue
from twisted.python import log
from twisted.python.filepath import FilePath
from twisted.python.procutils import which as twisted_which

from documint.errors import ExternalProcessError, NoSuchFile
from documint.util import defertee



def _checkStatus((out, err, code)):
    """
    Check that the exit status is C{0}.
    """
    return code == 0



def getProcessOutput(binary, args,
                     getProcessOutputAndValue=getProcessOutputAndValue,
                     checkStatus=_checkStatus):
    """
    Spawn a new process and return the standard output.

    @type  binary: C{str}
    @param binary: Path to the binary to spawn.

    @type  args: C{sequence} of C{str}
    @param args: Arguments to pass when spawning L{binary}.

    @param getProcessOutputAndValue: Callable to spawn a process and return
        C{(standard output, standard error, exit status)}. Defaults to
        L{twisted.internet.utils.getProcessOutputAndValue}

    @raise L{ExternalProcessError}: If the exit status is non-zero.

    @rtype: C{Deferred<str>}
    @return: Deferred firing with the standard output of L{binary}.
    """
    def _log((out, err, code)):
        name = FilePath(binary).basename()
        log.msg('%s command: %r' % (name, (binary, args)))

    def _check((out, err, code)):
        if not checkStatus((out, err, code)):
            raise ExternalProcessError(binary, args, code, (out, err))
        return out

    d = getProcessOutputAndValue(binary, args)
    d.addCallback(defertee, _log)
    d.addCallback(_check)
    return d



def sanitizePaths(paths):
    """
    Convert an iterable of C{FilePath}s to an iterable of C{str}.
    """
    for path in paths:
        if isinstance(path, (str, unicode)):
            warn(
                'paths should be twisted.python.filepath.FilePath not %r' % (
                    type(path),),
                DeprecationWarning, 2)
            path = FilePath(path)

        fn = path.path
        if isinstance(fn, unicode):
            fn = fn.encode('ascii')

        if not exists(fn):
            raise NoSuchFile(fn)

        yield fn



def which(binary, _which=twisted_which):
    """
    Locate the first binary in the path.

    @type  binary: C{str}
    @param binary: Name of the binary to find.

    @param _which: Callable taking one parameter, the name of the binary to
        find, returning a list of possibilities.

    @raise RuntimeError: If the binary could not be located.

    @return: Path of the binary.
    """
    paths = _which(binary)
    if paths:
        return paths[0]
    raise RuntimeError('binary %r not found' % (binary,))



from functools import partial

from documint.extproc.common import getProcessOutput, which



findFop = partial(which, 'fop')



def fop(inputFile, outputFile, configFile=None):
    """
    Run Apache FOP.

    @type  inputFile: L{twisted.python.filepath.FilePath}
    @param inputFile: Input file path.

    @type  outputFile: L{twisted.python.filepath.FilePath}
    @param outputFile: Output file path.

    @type  configFile: L{twisted.python.filepath.FilePath}
    @param configFile: Optional config file path.

    @raise L{ExternalProcessError}: If FOP fails to complete successfully.

    @rtype: C{Deferred<str>}
    @return: Deferred that fires with the byte data of the resulting PDF
        document.
    """
    def _readData(ignored):
        return outputFile.getContent()

    args = []
    if configFile is not None:
        args.extend(['-c', configFile.path])
    args.extend([inputFile.path, outputFile.path])
    d = getProcessOutput(findFop(), args)
    d.addCallback(_readData)
    return d



"""
I{Documint} utility functions.
"""
from StringIO import StringIO

from lxml import etree
from twisted.internet.defer import maybeDeferred



def embedStylesheets(markup, stylesheets, removeStylesheets=False):
    """
    Embed external stylesheets in I{XHTML} markup.

    @type  markup: L{str}
    @param markup: I{XHTML} markup UTF-8 byte data.

    @type  stylesheets: I{iterable} of L{str}
    @param stylesheets: Iterable of stylesheet UTF-8 byte data to embed.

    @rtype: L{str}
    @return: I{XHTML} UTF-8 byte data with embedded stylesheets.
    """
    tree = etree.parse(StringIO(markup))
    namespaces = {'xhtml': 'http://www.w3.org/1999/xhtml'}
    head = tree.xpath('//xhtml:head', namespaces=namespaces)[0]

    for stylesheet in stylesheets:
        e = etree.SubElement(
            head,
            '{http://www.w3.org/1999/xhtml}style',
            attrib=dict(type='text/css'))
        e.text = stylesheet.decode('utf-8')

    if removeStylesheets:
        stylesheetLinks = head.xpath(
            'xhtml:link[@rel="stylesheet"]', namespaces=namespaces)
        for e in stylesheetLinks:
            head.remove(e)

    return etree.tostring(tree, encoding='utf-8')


def defertee(result, func, *a, **kw):
    """
    Call C{func}, with positional and keyword arguments, as a side effect, and
    return C{result}.

    Useful in combination with C{Deferred.addCallback} when you wish to perform
    an operation in the callback chain but wish to retain the result for
    subsequent callbacks.
    """
    d = maybeDeferred(func, result, *a, **kw)
    d.addCallback(lambda ignored: result)
    return d



__all__ = ['embedStylesheets']



"""
Tests for L{documint.commands}.
"""
from twisted.internet.defer import succeed
from twisted.trial.unittest import TestCase

from documint.commands import Minter, SimpleMinter



class MinterTests(TestCase):
    """
    Tests for L{documint.commands.Minter}.
    """
    def test_render(self):
        """
        L{Minter.render}, a responder for L{documint.commands.Render}, invokes
        L{Minter.renderXHTML}.
        """
        minter = Minter()
        minter.renderXHTML = lambda *a: succeed((a, 'application/pdf'))
        d = minter.render('markup', ['css1', 'css2'])
        d.addCallback(
            self.assertEquals,
            {'data': ('markup', ['css1', 'css2']),
             'contentType': 'application/pdf'})
        return d



class SimpleMinterTests(TestCase):
    """
    Tests for L{documint.commands.SimpleMinter}.
    """
    def test_renderXHTML(self):
        """
        L{SimpleMinter.renderXHTML} invokes L{Minter.embedStylesheets}.
        """
        minter = SimpleMinter()
        minter.embedStylesheets = lambda *a: a
        d = minter.renderXHTML('markup', ['css1', 'css2'])
        d.addCallback(
            self.assertEquals, (('markup', ['css1', 'css2']), 'text/html'))
        return d



from setuptools import find_packages, setup



setup(
    name="Documint",
    version="15.5.0",
    url="https://github.com/fusionapp/documint",
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Framework :: Twisted',
        'Intended Audience :: Developers',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 2.7',
        'Topic :: Text Processing :: Markup :: HTML',
        ],
    packages=find_packages() + ['twisted.plugins'],
    install_requires=['Twisted', 'lxml'],
    include_package_data=True)
